#include <errno.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>

#include "capi.h"

const char *PROG_NAME = "capid";

static int send_two_parse_errors = 0;
static int skip_crlf = 0;

static void usage(void)
{
	fprintf(stderr, "Usage: %s [options]\n", PROG_NAME);
	fprintf(stderr, "\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "   -h               usage\n");
	fprintf(stderr, "   -p <portnum>     port number\n");
	fprintf(stderr, "   -t               send two errors when parsing fail\n");
	fprintf(stderr, "   -n 		     skip CRLF at the output lines ending\n");
	//fprintf(stderr, "   -s <socket path> ubus socket\n");
	//fprintf(stderr, "   -D               run as a daemon\n");
	//fprintf(stderr, "   -c <conf-file>   specify configuration file\n");
	//fprintf(stderr, "   -d               debug level; more 'd's mean more verbose\n");
	fprintf(stderr, "   -o <file>        log to file\n");
	//fprintf(stderr, "   -f               treat above file as fifo\n");
	fprintf(stderr, "\n");
}


struct capi *capi_init(uint16_t port)
{
	struct capi *ctx;
	struct sockaddr_in sa = {0};
	int reuse = 1;
	int flags;
	int s;


	ctx = calloc(1, sizeof(*ctx));
	if (!ctx)
		return NULL;

	INIT_LIST_HEAD(&ctx->clientlist);

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		goto out_error;

	flags = fcntl(s, F_GETFL, 0);
	if (flags != -1) {
		fcntl(s, F_SETFL, flags | O_NONBLOCK);
	}

	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) != 0) {
		fprintf(stderr, "reuseaddr error\n");
		close(s);
		goto out_error;
	}

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = INADDR_ANY;
	sa.sin_port = htons(port);
	if ((bind(s, (struct sockaddr *)&sa, sizeof(struct sockaddr_in))) != 0) {
		fprintf(stderr, "bind() error\n");
		close(s);
		goto out_error;
	}

	if (listen(s, CAPI_BACKLOG) != 0) {
		fprintf(stderr, "bind() error\n");
		close(s);
		goto out_error;
	}

	fprintf(stderr, "listening on port %hu\n", port);
	ctx->sk = s;
	ctx->nfds = s;
	ctx->port = port;
	FD_ZERO(&ctx->rfds);
	FD_ZERO(&ctx->wfds);
	FD_SET(ctx->nfds, &ctx->rfds);

	return ctx;

out_error:
	free(ctx);
	return NULL;
}

void capi_exit(struct capi *ctx)
{
	struct capi_client *cli, *tmp;

	list_for_each_entry_safe(cli, tmp, &ctx->clientlist, list) {
		list_del(&cli->list);
		//TODO:free cmdlist
		free(cli);
	}

	close(ctx->sk);
	free(ctx);
}

int capi_accept_client(struct capi *ctx)
{
	socklen_t cslen = sizeof(struct sockaddr_in);
	struct sockaddr_in cs;
	struct capi_client *cli;
	int flags;


	fprintf(stderr, "client connected\n");
	cli = calloc(1, sizeof(*cli));
	if (!cli)
		return -1;

	cli->sk = accept(ctx->sk, (struct sockaddr *)&cs, &cslen);
	if (cli->sk < 0) {
		fprintf(stderr, "accept() error! (%d)\n", errno);
		goto out_error;
	}

	ctx->nfds = cli->sk;
	flags = fcntl(cli->sk, F_GETFL, 0);
	if (flags != -1) {
		fcntl(cli->sk, F_SETFL, flags | O_NONBLOCK);
	}

	FD_SET(cli->sk, &ctx->rfds);
	list_add_tail(&cli->list, &ctx->clientlist);

	return 0;

out_error:
	free(cli);
	return -1;
}

int capi_send_response(struct capi *ctx, struct capi_client *cli,
		       enum capi_state s, char *response)
{
	char resp[CAPI_RESP_SIZE] = {0};
	size_t len = 0;
	int ret;

	fprintf(stderr, "sending ACK ===> (state = %d)\n", s);

	switch (s) {
	case CAPI_INVALID:
		sprintf(resp, "status,%s", "INVALID");
		break;
	case CAPI_RUNNING:
		sprintf(resp, "status,%s", "RUNNING");
		break;
	case CAPI_ERROR:
		sprintf(resp, "status,%s", "ERROR");
		break;
	case CAPI_COMPLETE:
		sprintf(resp, "status,%s", "COMPLETE");
		break;
	default:
		return -1;
	}

	len = strlen(resp);
	if (response)
		snprintf(resp + len, CAPI_RESP_SIZE - len - 1,
			 ",%s%s", response, skip_crlf ? "" : "\r\n");
	else
		snprintf(resp + len, CAPI_RESP_SIZE - len - 1,
			 "%s", skip_crlf ? "" : "\r\n");

	len = strlen(resp);
	ret = send(cli->sk, resp, len, 0);
	if (ret < 0) {
		if (errno == EAGAIN || errno == EWOULDBLOCK) {
			fprintf(stderr, "%s: %d\n", __func__, __LINE__);
			FD_SET(cli->sk, &ctx->wfds);
			return -1;
		}
	}

	fprintf(stderr, "%s: len = %zu  <===\n", __func__, len);
	return 0;
}

int capi_disconnect_client(struct capi *ctx, struct capi_client *cli)
{
	fprintf(stderr, "client disconnected\n");

	FD_CLR(cli->sk, &ctx->rfds);
	close(cli->sk);
	return 0;
}

int capi_recv(struct capi *ctx, struct capi_client *cli)
{
	char buf[CAPI_CMD_SIZE];
	char *ptr, *saveptr;
	int ret, len;


	ret = recv(cli->sk, buf, sizeof(buf) - 1, 0);
	if (!ret)
		return capi_disconnect_client(ctx, cli);

	if (ret < 0) {
		if (errno == EAGAIN || errno == EWOULDBLOCK)
			goto out;

		fprintf(stderr, "recv() error! (%d)\n", errno);
		return -1;
	}

	buf[ret] = '\0';
	fprintf(stderr, "Recvd: '%s'\n", buf);

	ptr = buf;
	for (;;) {
		char response[CAPI_RESP_SIZE] = {0};
		char *line, *line2;

		line = strtok_r(ptr, "\n", &saveptr);
		if (line == NULL)
			break;
		ptr = saveptr;

		len = strlen(line);
		if (len == 0)
			continue;

		fprintf(stderr, "%s: line = '%s' len = %d\n", __func__, line, len);

		while (len > 0 && (isspace(line[len-1]) || line[len-1] == '\n' || line[len-1] == '\r'))
			line[--len] = '\0';

		if (line[0] == '\0')
			continue;

		line2 = strdup(line);
		ret = capi_validate_cmd(ctx, cli, line2);
		free(line2);

		if (!ret) {
			capi_send_response(ctx, cli, CAPI_RUNNING, NULL);
			ret = capi_run_cmd(ctx, cli, line, response);
			capi_send_response(ctx, cli, !ret ? CAPI_COMPLETE : CAPI_ERROR, response);
		} else {
			if (send_two_parse_errors) {
				/* UCC might expect response with second status */
				capi_send_response(ctx, cli, CAPI_INVALID, NULL);
				capi_send_response(ctx, cli, CAPI_ERROR, NULL);
			} else {
				snprintf(response, CAPI_RESP_SIZE, "errorCode,%d", -ret);
				capi_send_response(ctx, cli, CAPI_INVALID, response);
			}
		}
	}

out:
	FD_SET(cli->sk, &ctx->rfds);

	return 0;
}

int capi_run(uint16_t port)
{
	struct capi *ctx;
	struct capi_client *cli;
	int ret;


	ctx = capi_init(port);
	if (!ctx)
		return -1;

	//capi_init_easymesh(ctx);	//TODO: TODO:
	capi_call(easymesh);

	for (;;) {
		fd_set r, w;

		r = ctx->rfds;
		w = ctx->wfds;
		ret = select(ctx->nfds + 1, &r, &w, NULL, NULL);
		if (ret < 0) {
			fprintf(stderr, "select() error! (%d)\n", errno);
			perror("exit");
			break;
		}

		if (FD_ISSET(ctx->sk, &r)) {
			ret = capi_accept_client(ctx);
		} else {
			list_for_each_entry(cli, &ctx->clientlist, list) {
				if (FD_ISSET(cli->sk, &r))
					ret |= capi_recv(ctx, cli);
			}
		}
	}

	capi_exit(ctx);

	return 0;
}

int main(int argc, char **argv)
{
	int ch;
	uint16_t portnum = 9119;


	while ((ch = getopt(argc, argv, "htndp:o:")) != -1) {
		switch (ch) {
		case 'h':
			usage();
			exit(0);
		case 'p':
			portnum = strtoul(optarg, NULL, 10);
			break;
		case 't':
			send_two_parse_errors = 1;
			break;
		case 'n':
			skip_crlf = 1;
			break;
		case 'd':
			setenv("CAPI_DEBUG", "1", 1);
			break;
		case 'o':
			setenv("CAPI_LOGFILE", optarg, 1);
			break;
		default:
			break;
		}
	}

	argc -= optind;
	argv += optind;

	capi_run(portnum);

}
